#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

typedef uint8_t		u8;
typedef	uint16_t	u16;
typedef uint32_t	u32;
typedef int8_t		s8;
typedef int16_t		s16;

typedef struct {
	s16	scale;
	s16	yn1;
	s16	yn0;
	u8	data[64];
} PACKET;

typedef struct {
	u32	magic;
	u32	length;
	u16	format_tag;
	u16	channels;
	u32	sample_rate;
	u32	byte_rate;
	u16	frame_size;
	u16	bits;
} FMT;

typedef struct {
	u32	magic;
	u32	length;
	s16	data[0];
} DATA;

typedef struct {
	u32	magic;
	u32	filesize;
	u32	wave;
	FMT	fmt;
	DATA	data;
} WAVE;

static void readall(int fd, void* buf, size_t size)
{
	ssize_t rd;
	size_t sz = size;
	u8* p = (u8*) buf;
	while(sz != 0 && (rd = read(fd, p, sz)) != 0) {
		if(rd == -1) {
			perror("read");
			exit(1);
		}
		p += rd;
		sz -= rd;
	}
}

static void writeall(int fd, void* buf, size_t size)
{
	ssize_t wr;
	size_t sz = size;
	u8* p = (u8*) buf;
	while(sz != 0 && (wr = write(fd, p, sz)) != 0) {
		if(wr == -1) {
			perror("write");
			exit(1);
		}
		p += wr;
		sz -= wr;
	}
}

static inline s16 clamp16(int val)
{
	if(val > 32767)
		return 32767;
	if(val < -32768)
		return -32768;
	return (s16) val;
}

s16 decode(PACKET* pkt, s16* outbuf)
{
	int scale = pkt->scale;

	/* not using the history avoids clicks since our predictor is wrong */
#if 0
	int yn0 = pkt->yn0;
	int yn1 = pkt->yn1;
#else
	static int yn0 = 0;
	static int yn1 = 0;
#endif

	for(unsigned int i = 0; i < 256; i++) {
		unsigned int j = i % 4; /* bit group */

		s8 byte = pkt->data[i / 4];
		s8 bits = (byte << (2 * j)) & 0xC0;
		bits >>= 6; /* get 2bit signed */

		int value = bits * scale + yn0 * 61 / 32 - yn1 * 15 / 16;

		yn1 = yn0;
		yn0 = value;

		outbuf[i] = clamp16(value);
	}

	return yn0;
}

int main(int argc, char** argv)
{
	if(argc != 3) {
		printf("Usage: %s in.bin out.wav\n", *argv);
		return 1;
	}

	const char* infile = argv[1];
	const char* outfile = argv[2];

	struct stat buf;
	if(stat(infile, &buf) == -1) {
		perror("stat");
		return 1;
	}

	size_t blocks = buf.st_size / 70;
	printf("%ld blocks\n", blocks);

	int fd = open(infile, O_RDONLY);
	if(fd == -1) {
		perror("open");
		return 1;
	}

	int fdo = open(outfile, O_WRONLY | O_CREAT, 0644);
	if(fdo == -1) {
		perror("open");
		return 1;
	}

#define	SAMPLES_PER_BLOCK	256
	WAVE wav;
	wav.magic = 0x46464952;
	wav.filesize = blocks * SAMPLES_PER_BLOCK * 2 + sizeof(WAVE) - 8;
	wav.wave = 0x45564157;
	wav.fmt.magic = 0x20746D66;
	wav.fmt.length = sizeof(wav.fmt) - 8;
	wav.fmt.format_tag = 1;
	wav.fmt.channels = 1;
	wav.fmt.sample_rate = 16000;
	wav.fmt.byte_rate = wav.fmt.sample_rate * wav.fmt.channels * 2;
	wav.fmt.frame_size = 2;
	wav.fmt.bits = 16;
	wav.data.magic = 0x61746164;
	wav.data.length = wav.filesize - sizeof(WAVE);
	writeall(fdo, &wav, sizeof(wav));

	for(size_t i = 0; i < blocks; i++) {
		PACKET pkt;
		s16 outbuf[SAMPLES_PER_BLOCK];

		readall(fd, &pkt, sizeof(PACKET));

		decode(&pkt, outbuf);

		writeall(fdo, outbuf, SAMPLES_PER_BLOCK * 2);
	}

	close(fdo);
	close(fd);
}
